from random import choice, randint, uniform
from math2 import Vector, Segment, Shape
from sets import Set
from render import WallQuad, FloorQuad

def pairs(l):
	return zip(l[::2], l[1::2])
	#for x in range(0, len(l), 2):
	#	yield (l[x], l[x+1])

def segments_sort(l):
	if len(l) < 2:
		return

	ref = l[0]
	l.sort(lambda a, b: cmp(a.subv(ref).length(), b.subv(ref).length()))


class Doorway (object):
	ROOM = 1
	CORRIDOR = 2
	def __init__(self, loc, heading, dtype, feature):
		self.loc = loc
		self.heading = heading
		self.dtype = dtype
		self.feature = feature

class Feature (Shape):
	def __init__(self, data):
		Shape.__init__(self, data)
		self.connections = {}

	def connect(self, location, other):
		self.connections[location] = other

class Corridor (Feature):
	def __init__(self, start, end, heading):
		Feature.__init__(self, [Segment(start, end)])
		self.ends = [start, end]
		self.start = start
		self.end = end
		self.heading = heading
	
	def length(self):
		return self.segments[0].length()

	def connect(self, location, other):
		# Allow rooms to replace connections to corridors, but not
		# otherwise.
		if location in self.connections:
			assert(isinstance(other, Room))
		Feature.connect(self, location, other)
	
	def __repr__(self):
		return '<Corridor(%.2f, %.2f, %.2f, %.2f)>' % (
			self.start.x, self.start.y, self.end.x, self.end.y)

	def build(self, walltex, floortex, scale=0.25):
		segments = []
		shapes = []

		start = self.start
		end = self.end

		head = self.heading
		unit = head.mul(scale)

		su = head.mul(0.2)
		sul = su.left()
		sur = su.right()

		# Side walls.
		wls = start.addv(unit.left())
		wle = end.addv(unit.left())
		wrs = end.addv(unit.right())
		wre = start.addv(unit.right())

		wl_h = head
		wr_h = head.mul(-1)

		# Shape corners.
		wols = wls.addv(sul)
		wole = wle.addv(sul)
		wors = wrs.addv(sur)
		wore = wre.addv(sur)

		start_feature = self.connections.get(start, None)
		# Make corridor join seamlessly with less segments.
		if start_feature:
			if isinstance(start_feature, Corridor):
				if head.right() == start_feature.heading:
					# Left is unit shorter, right is unit
					# longer
					wls = wls.addv(unit)
					wols = wls.addv(sul).addv(su)
					wre = wre.subv(unit)
					wore = wre.addv(sur).subv(su)
				elif head.left() == start_feature.heading:
					# Left is unit longer, right is unit
					# shorter.
					wls = wls.subv(unit)
					wols = wls.addv(sul).subv(su)
					wre = wre.addv(unit)
					wore = wre.addv(sur).addv(su)
				# Else we are collinear, no modifications.

			elif isinstance(start_feature, Room):
				# Both are unit shorter.
				wls = wls.addv(unit)
				wols = wls.addv(sul).addv(su)
				wre = wre.addv(unit)
				wore = wre.addv(sur).addv(su)

		end_feature = self.connections.get(end, None)
		if end_feature is not None:
			if isinstance(end_feature, Corridor):
				if end in end_feature.ends:
					efhead = end_feature.heading
					if end == end_feature.end:
						efhead = efhead.mul(-1)

					if head.left() == efhead:
						# Right wall is unit longer, 
						# left is unit shorter.
						wle = wle.subv(unit)
						wole = wle.addv(sul).subv(su)
						wrs = wrs.addv(unit)
						wors = wrs.addv(sur).addv(su)
					elif head.right() == efhead:
						# Right wall is unit shorter, 
						# right is unit longer.
						wle = wle.addv(unit)
						wole = wle.addv(sul).addv(su)
						wrs = wrs.subv(unit)
						wors = wrs.addv(sur).subv(su)
				else:
					# We're running into the middle of
					# other corridor, both walls are
					# shorter.
					wle = wle.subv(unit)
					wole = wle.addv(sul).subv(su)
					wrs = wrs.subv(unit)
					wors = wrs.addv(sur).subv(su)

			elif isinstance(end_feature, Room):
				# Both are unit shorter.
				wle = wle.subv(unit)
				wole = wle.addv(sul).subv(su)
				wrs = wrs.subv(unit)
				wors = wrs.addv(sur).subv(su)
		else:
			seg = Segment(wle, wrs)
			segments.append(seg)
			wole = wole.addv(su)
			wors = wors.addv(su)
			seg.quad = WallQuad(walltex, [wle, wrs, wors, wole])
			shapes.append(seg.quad)

		wl = [wls, wle]
		wr = [wrs, wre]

		# Make floor.
		shapes.append(FloorQuad(floortex, [wls, wle, wrs, wre]))
		
		# T-intersections.
		for point, feature in self.connections.items():
			if point in self.ends:
				continue
			
			wp = point.subv(feature.heading.mul(scale))
			wh = feature.heading.left()
			whs = wh.mul(scale)
			
			c1 = wp.subv(whs)
			c2 = wp.addv(whs)

			if wh == wl_h:
				wl.append(c1)
				wl.append(c2)
			elif wh == wr_h:
				wr.append(c1)
				wr.append(c2)
			else:
				raise 'InternalError'

		segments_sort(wl)
		segments_sort(wr)

		for p1, p2 in pairs(wl):
			seg = Segment(p1, p2)
			segments.append(seg)

			if p1 == wls:
				p1o = wols
			else:
				p1o = p1.addv(sul).addv(su)
			if p2 == wle:
				p2o = wole
			else:
				p2o = p2.addv(sul).subv(su)

			seg.quad = WallQuad(walltex, [p1, p2, p2o, p1o])
			shapes.append(seg.quad)

		for p1, p2 in pairs(wr):
			seg = Segment(p1, p2)
			segments.append(seg)

			if p1 == wrs:
				p1o = wors
			else:
				p1o = p1.addv(sur).subv(su)

			if p2 == wre:
				p2o = wore
			else:
				p2o = p2.addv(sur).addv(su)

			seg.quad = WallQuad(walltex, [p1, p2, p2o, p1o])
			shapes.append(seg.quad)

		return segments, shapes

class Room (Feature):
	def __init__(self, c, s):
		hs = s.mul(0.5)
		p1 = self.p1 = c.subv(hs)
		p2 = self.p2 = c.addv(hs)
		Feature.__init__(self, self.rectangle_segments(p1.x, p1.y, s.x, s.y))

		self.connections = Set()
	
	def connect(self, location, other):
		self.connections.add((location, other))
	
	def build(self, walltex, floortex, scale=0.25):
		walls = {}
		corners = []
		shapes = []

		for wall in self.segments:
			head = wall.p2.subv(wall.p1).normal()
			unit = head.mul(scale)

			p1 = wall.p1.addv(unit.left()).subv(unit)
			p2 = wall.p2.addv(unit.left()).addv(unit)

			corners.append(p1)
			walls[head] = [p1, p2]

		shapes.append(FloorQuad(floortex, corners))

		for point, feature in self.connections:
			assert(isinstance(feature, Corridor))

			fh = feature.heading

			if point == feature.start:
				whead = fh.right()
				fhs = fh.mul(scale)
			elif point == feature.end:
				whead = fh.left()
				fhs = fh.mul(-scale)

			whs = whead.mul(scale)

			c1 = point.subv(whs).addv(fhs)
			c2 = point.addv(whs).addv(fhs)

			walls[whead].append(c1)
			walls[whead].append(c2)

		segments = []

		for whead, points in walls.items():
			segments_sort(points)

			su = whead.mul(0.2)
			sul = su.left()

			ws = points[0]
			we = points[-1]

			for p1, p2 in pairs(points):
				if p2.subv(p1).length() == 0:
					continue
				
				p1o = p1.addv(sul)
				p2o = p2.addv(sul)

				if p1 == ws:
					p1o = p1o.subv(su)
				else:
					p1o = p1o.addv(su)

				if p2 == we:
					p2o = p2o.addv(su)
				else:
					p2o = p2o.subv(su)

				seg = Segment(p1, p2)
				segments.append(seg)

				seg.quad = WallQuad(walltex, [p1, p2, p2o, p1o])
				shapes.append(seg.quad)
	
		return segments, shapes
			

class Digger (object):
	"""Generate nethack style mazes.

	http://www.jimrandomh.org/rldev/digging_features/index.html
	"""

	directions = [Vector(1, 0), Vector(0, 1), Vector(-1, 0), Vector(0, -1)]	
	
	def __init__(self, minrooms=5, maxrooms=6):
		self.minrooms = minrooms
		self.maxrooms = maxrooms
		self.features = []

	def randvect(self, v):
		x, y = 0, 0
		if int(v.x):
			x = randint(0, abs(int(v.x)))
			if v.x < 0:
				x = -x
		if int(v.y):
			y = randint(0, abs(int(v.y)))
			if v.y < 0:
				y = -y
			
		return Vector(x, y)

	def randsize(self):
		return Vector(randint(2, 4), randint(2, 4))

	def generate(self, cb=None):
		# Clear.
		self.doorways = {}
		self.features = []
		self.rooms = 0
		# Room is centered at (0, 0), so to avoid fractional corners,
		# use even size.
		self.add_room(Vector(0, 0), self.randsize().mul(2), mindoors=2)

		if cb: cb(self)

		while self.doorways:
			door_loc = choice(self.doorways.keys())
			door = self.doorways[door_loc]
			del self.doorways[door_loc]
			self.make_feature(door)
			if cb: cb(self)
	
	def make_feature(self, door):
		loc = door.loc
		head = door.heading
		if door.dtype == door.ROOM or uniform(0, 1) < 0.4:
			if door.dtype == door.CORRIDOR:
				head = choice([head, head.left(), head.right()])

			return self.make_corridor(loc, head, door.feature)
		else:
			return self.make_room(loc, head, door.feature)
	
	def make_room(self, entrance, heading, connect):
		size = self.randsize()
		# Find room corner to the left of the entrance.
		offset = randint(0, size.x)
		corner = entrance.addv(heading.left().mul(offset))
		# Rotate size with room.
		nsize = heading.right().mul(size.x).addv(heading.mul(size.y)).abs()
		# From corner to the center.
		delta = heading.right().mul(size.x/2).addv(heading.mul(size.y/2))
		center = corner.addv(delta)

		self.add_room(center, nsize, (entrance, connect))

	def add_room(self, center, size, connect=None, mindoors=0):
		room = Room(center, size)

		collisions = self.find_overlap(room)
		to_connect = []
		for point, feature in collisions:
			# We can't connect or overlap with rooms.
			if isinstance(feature, Room):
				return

			# We only allow intersection with corridors when it ends on
			# our wall.
			elif isinstance(feature, Corridor):
				if point in feature.ends:
					# Multiple touch is not allowed.
					if (point, feature) in room.connections:
						return
					# We're touching, thus connect.
					to_connect.append((point, feature))
					continue
				else:
					return

		for point, feature in to_connect:
			room.connect(point, feature)
			feature.connect(point, room)

		self.rooms += 1
		self.features.append(room)

		#if connect:
			#location, feature = connect
			#room.connect(location, feature)
			#feature.connect(location, room)
		

		if self.rooms < self.minrooms:
			mindoors = 2
		if self.rooms > self.maxrooms:
			# Enough is enough.
			return

		hsize = size.mul(0.5)
		directions = self.directions[:]

		for x in range(randint(mindoors, 4)):
			head = choice(directions)
			directions.remove(head)
			# center of wall.
			corner = center.addv(hsize.scale(head)).addv(
					hsize.scale(head.left()))
			loc = corner.addv(self.randvect(size.scale(head.right())))
			self.doorways[loc] = Doorway(loc, head, Doorway.ROOM, room)


	def make_corridor(self, loc, heading, connect):
		length = randint(3, 5)
		endpoint = loc.addv(heading.mul(length))
		corr = Corridor(loc, endpoint, heading)

		collisions = [(v, o) for v, o in self.find_overlap(corr) if v != loc]
		for v, o in collisions:
			if not isinstance(v, Vector):
				return

		if collisions:
			mincoll = [collisions[0]]
			mindist = mincoll[0][0].subv(loc).length()
			for value, other in collisions[1:]:
				# Do not allow common segments with other features.
				dist = value.subv(loc).length()
				if dist < mindist:
					mincoll = [(value, other)]
					mindist = dist
				elif dist == mindist:
					mincoll.append((value, other))

		if collisions:
			if len(mincoll) > 1:
				for endpoint, end_feature in mincoll:
					if isinstance(end_feature, Room):
						break
				# If both collisions are corridors, then we
				# ran straight between the joint. This can
				# be handled, but won't.
				if not isinstance(end_feature, Room):
					return

			else:
				endpoint, end_feature = mincoll[0]
			
			corr = Corridor(loc, endpoint, heading)

			if corr.length() < 2:
				return

			self.features.append(corr)
			corr.connect(endpoint, end_feature)
			end_feature.connect(endpoint, corr)

			if endpoint in self.doorways:
				del self.doorways[endpoint]

		else:
			self.features.append(corr)
			self.doorways[endpoint] = Doorway(endpoint, heading, Doorway.CORRIDOR, corr)

		# Interconnect
		corr.connect(loc, connect)
		connect.connect(loc, corr)

	def find_overlap(self, this):
		collisions = []
		for other in self.features:
			coll = this.contacts(other)
			for value, segment in coll:
				if (value, other) not in collisions:
					collisions.append((value, other))
		return collisions

	def build(self, walltex, floortex, scale=0.25):
		segments = []
		shapes = []
		for f in self.features:
			fseg, fshp = f.build(walltex, floortex, scale)
			segments.extend(fseg)
			shapes.extend(fshp)

		#self.scale = scale = 0.5 / scale
		#segments = [s.scale(scale) for s in segments]
		#shapes = [s.scale(scale) for s in shapes]

		return segments, shapes

if __name__ == '__main__':
	import pygame
	from pygame.locals import *
	from render import Texture
	from OpenGL.GL import *

	pygame.init()
	disp = pygame.display.set_mode((600, 600), DOUBLEBUF | OPENGL)
	lookat = Vector(0, 0)

	maze = Digger(5, 10)
	scale = 0.5

	#def draw_seg(seg, color):
	#	p1 = seg.p1.mul(scale).addv(cs).tupi()
	#	p2 = seg.p2.mul(scale).addv(cs).tupi()
	#	pygame.draw.line(disp, color, p1, p2)


	#def callback(maze):
	#	feature = maze.features[-1]
	#	if isinstance(feature, Corridor):
	#		color = (255, 0, 0)
	#	elif isinstance(feature, Room):
	#		color = (255, 255, 255)

	#	for seg in feature.segments:
	#		draw_seg(seg, color)

	#	pygame.display.update()

	wtex = Texture('data/lab-wall.png')
	ftex = Texture('data/lab-floor2.png')
	wtex.load()
	ftex.load()

	glEnable(GL_TEXTURE_2D)

	maze.generate()
	segments, shapes = maze.build(wtex, ftex)
	print 'Done, map edges:', len(segments)
	clock = pygame.time.Clock()

	k_state = {}

	glNewList(1, GL_COMPILE)
	for poly in shapes:
#			points = poly.vpoints[:]
#			points.append(points[0])
#			for p1, p2 in zip(points[:-1], points[1:]):
#				seg = Segment(p1, p2)
#				draw_seg(seg, (0, 255, 255))
		poly.render()
	glEndList()

	while 1:
		for ev in pygame.event.get():
			if ev.type == KEYDOWN:
				if ev.key == K_SPACE:
					disp.fill(0)
					maze.generate(callback)
					segments, shapes = maze.build(wtex, ftex)
					print 'Done, map edges: ', len(segments)
				if ev.key == K_ESCAPE:
					break

			if ev.type in [KEYUP, KEYDOWN]:
				k_state[ev.key] = ev.type == KEYDOWN

		if k_state.get(K_LEFT, 0):
			lookat = lookat.addi(0.05, 0)
		if k_state.get(K_RIGHT, 0):
			lookat = lookat.addi(-0.05, 0)
		if k_state.get(K_UP, 0):
			lookat = lookat.addi(0, 0.05)
		if k_state.get(K_DOWN, 0):
			lookat = lookat.addi(0, -0.05)
		if k_state.get(K_q, 0):
			scale += 0.01
		if k_state.get(K_a, 0):
			scale -= 0.01

		glClear(GL_COLOR_BUFFER_BIT)
		glLoadIdentity()
		glScalef(scale, scale, 1)
		glTranslatef(lookat.x, lookat.y, 0)

		glCallList(1)	
		
		pygame.display.flip()
		clock.tick(25)
	
