from quadtree import Quadtree
from render import VisibleShape
from math2 import Shape, Segment
from OpenGL.GL import *

class ShapeLevel(object):
	def __init__(self, engine, shape, entities):
		self.engine = engine
		self.shape = shape
		self.index = Quadtree()
		self.index.add_shape(shape)

		self.entities = entities

	def move(self, position, rotation):
		self.shape.move(position, rotation)

	def collide(self, other):
		bbox = other.bbox()
		segs = self.index.likely_intersection(bbox)
		self.last_coll_segments = len(segs)
		return other.collide(Shape(segs))
	
	def check_beam(self, ent, beam, player):
		bs = Segment(ent.loc, beam)
		bss = Shape([bs])
		
		if player.shape.contacts(bss):
			mindist = player.drive.position.subv(ent.loc).length()
			bs = Segment(ent.loc, player.drive.position)
			player.take_hit()

		return bs

	def step(self, t, player):
		self.beams = []
		for e in self.entities:
			e.step(t)
			if e.beam:
				beam = self.check_beam(e, e.beam, player)
				self.beams.append(beam)

	def compile(self):
		self.glist = glGenLists(1)
		glNewList(self.glist, GL_COMPILE)
		for quad in self.shape.real_quads:
			quad.render()
		glEndList()
	
	def render(self):
		glCallList(self.glist)
		for e in self.entities:
			e.render()
		glPushAttrib(GL_CURRENT_BIT | GL_LINE_BIT)
		glLineWidth(5)
		glBegin(GL_LINES)
		glColor3f(1, 0.5, 0.5)
		for b in self.beams:
			glVertex2f(b.p1.x, b.p1.y)
			glVertex2f(b.p2.x, b.p2.y)
		glEnd()
		glPopAttrib()

if __name__ == '__main__':
	from digger import Digger
	maze = Digger()
	maze.generate()
	maze_segs, maze_shapes = maze.build(None, None)
	maze_shape = VisibleShape(maze_segs, maze_shapes)
	level = ShapeLevel(maze_shape)
	print len(maze_shape)
