from sets import Set

class Quadtree (object):
	"""Not a real quadtree, just a peformance hack."""

	def __init__(self):
		self.data = {}
		self.total = 0
	
	def get_bounds(self, segment):
		if isinstance(segment, tuple):
			p1x, p1y, p2x, p2y = tuple(map(int, segment))
		else:
			p1x = int(segment.p1.x)
			p1y = int(segment.p1.y)
			p2x = int(segment.p2.x)
			p2y = int(segment.p2.y)

		return (min(p1x, p2x)-1, min(p1y, p2y)-1, max(p1x, p2x)+1, max(p1y, p2y)+1)

	def add_segment(self, segment, data=None):
		p1x, p1y, p2x, p2y = self.get_bounds(segment)

		if not data:
			data = segment

		for y in range(p1y, p2y):
			for x in range(p1x, p2x):
				block = self.data.setdefault((x, y), Set())
				block.add(data)

		self.total += 1

	def add_segments(self, segments):
		for seg in segments:
			self.add_segment(seg)

	def add_shape(self, shape):
		self.add_segments(shape.real_segments)
	
	def likely_intersection(self, segment):
		p1x, p1y, p2x, p2y = self.get_bounds(segment)
		ret = Set()
		empty = Set()
		for y in range(p1y, p2y):
			for x in range(p1x, p2x):
				ret = ret.union(self.data.get((x, y), empty))
		
		return ret
	
if __name__ == '__main__':
	from digger import Digger

	maze = Digger(minrooms=10, maxrooms=15)

	print 'Doing maze.'
	maze.generate()
	segs = maze.build()

	print 'Doing quadtree.'
	qt = Quadtree()
	qt.add_segments(segs)

	print 'Regions', len(qt.data)

	blk_seg = map(len, qt.data.values())
	print 'Avg segs', sum(blk_seg) / float(len(qt.data))
	print 'Max segs', max(blk_seg)
	print 'Min segs', min(blk_seg)
	print 'Total refs', sum(blk_seg)
	print 'Total segs', qt.total
	
	import pygame
	from pygame.locals import *
	from math2 import Vector, Segment
	pygame.init()
	disp = pygame.display.set_mode((600, 600))
	clock = pygame.time.Clock()

	sc = Vector(300, 300)
	scale = 8
	
	def draw_seg(seg, color, surf=disp):
		p1 = seg.p1.mul(scale).addv(sc).tupi()
		p2 = seg.p2.mul(scale).addv(sc).tupi()
		pygame.draw.line(surf, color, p1, p2)

	level = pygame.Surface((600, 600))
	level.convert()
	for seg in segs:
		draw_seg(seg, (0, 255, 0), level)

	font = pygame.font.Font(pygame.font.get_default_font(), 12)

	box = Vector(0.5, 0.5)
	mouse = Vector(0, 0)

	while 1:
		for ev in pygame.event.get():
			if ev.type == KEYDOWN and ev.key == K_ESCAPE:
				break
			if ev.type == MOUSEMOTION:
				mouse = Vector(*ev.pos)
				mouse = mouse.subv(sc).mul(1./scale)

		disp.fill(0)
		disp.blit(level, (0, 0))

		sp1 = mouse.subv(box)
		sp2 = mouse.addv(box)
		mseg = Segment(sp1, sp2)

		tocheck = qt.likely_intersection(mseg)
		label = font.render('Check: %d at %.2fx%2f' % (len(tocheck), mouse.x, mouse.y), 
					1, (255, 255, 255))
		disp.blit(label, (0, 0))

		for seg in tocheck:
			draw_seg(seg, (255, 0, 0))

		draw_seg(mseg, (255, 255, 255))

		pygame.display.update()
		clock.tick(50)
