from digger import Digger
from backtracker import Backtracker
from level import ShapeLevel
from render import Texture, VisibleShape, Sprite
from combine import combine_shapes
from random import randint, choice, uniform
from math2 import Vector
from sets import Set
from turret import WallTurret

class TheLevelGenerator(object):
	def __init__(self, textures):
		self.textures = textures
	
	def digger_factory(self):
		minrooms = randint(3, 4)
		maxrooms = minrooms + randint(2, 3)
		return Digger(minrooms, maxrooms)

	def backtracker_factory(self):
		w = randint(10, 20)
		h = randint(10, 20)
		steps = w * h / 25
		return Backtracker(w, h, steps)

	def get_texture_set(self, what):
		lw = 'lab-wall.png'
		lf1 = 'lab-floor.png'
		lf2 = 'lab-floor2.png'
		cw = 'cave-wall.png'
		cf = 'cave-floor.png'
		
		textures = {
			Digger: { 
				'wall': [ lw ],
				'floor': [ lf1, lf2 ]
			},
			Backtracker: {
				'wall': [ cw ],
				'floor': [ cf ]
			}
		}

		wtex = self.textures[choice(textures[what]['wall'])]
		ftex = self.textures[choice(textures[what]['floor'])]

		return wtex, ftex

	def create_the_level(self, engine):
		# Shape combiner suck and suck more as I test it. :((((
		#factories = [ self.digger_factory,
		#	self.backtracker_factory]

		#levels = randint(3, 5)
		#dir = [Vector(-1, 0), Vector(0, -1), Vector(1, 0), Vector(0, 1)]
		#the_shape = None

		#while levels:
		#	# Player starts in the lab.
		#	if the_shape is not None:
		#		generator = choice(factories)()
		#	else:
		#		print 'No shape, starting with digger.'
		#		generator = self.digger_factory()
		#	try:
		#		generator.generate()
		#	except AssertionError:
		#		# There's some rare bug in Digger.generate(),
		#		# so to avoid spoinling game experience,
		#		# just try again. Working around bugs is bad. :(
		#		continue

		#	wtex, ftex = self.get_texture_set(type(generator))
		#	segments, shapes = generator.build(wtex, ftex)
		#	new_shape = VisibleShape(segments, shapes)
		#
		#	head = choice(dir)
		#	if the_shape is not None:
		#		print 'Combining.'
		#		the_shape = combine_shapes(
		#			the_shape, new_shape, head, ftex)
		#	else:
		#		the_shape = new_shape
		#	
		#	levels -= 1

		

		generator = self.digger_factory()
		generator.generate()
		wtex, ftex = self.get_texture_set(type(generator))
		segments, shapes = generator.build(wtex, ftex)
		the_shape = VisibleShape(segments, shapes)

		entities = []
		s = Sprite(0.25, 0.05, [self.textures['wall-turret.png']])
		s.load()

		walls_avail = the_shape.segments[:]
		nturrets = uniform(0.3, 0.5) * len(walls_avail)
		for count in range(int(nturrets)):
			wall = choice(walls_avail)
			walls_avail.remove(wall)
			offset = uniform(0.2, 0.8)
			wvec = wall.p2.subv(wall.p1)
			loc = wvec.mul(offset).addv(wall.p1)
			tur = WallTurret(engine, loc, wvec.normal().right(), s)
			entities.append(tur)

		level = ShapeLevel(engine, the_shape, entities)
		return level

if __name__ == '__main__':
	import pygame
	from render import TextureManager
	pygame.init()
	
	textures = TextureManager()
	tlg = TheLevelGenerator(textures)

	the_level = tlg.create_the_level()
	print 'Quads', len(the_level.shape.quads)
	print 'Segments', len(the_level.shape.segments)
	print 'Textures'
	print '\n'.join(['\t%s' % tex for tex in textures.textures])

